from typing import Optional, Callable

import panel
from atap_corpus._types import TCorpora
from atap_corpus.corpus.corpus import DataFrameCorpus
from panel.theme import Fast
from panel.viewable import Viewer, Viewable

from atap_corpus_loader.controller import Controller
from atap_corpus_loader.controller.events import EventType
from atap_corpus_loader.view import ViewWrapperWidget

panel.extension(notifications=True, design=Fast)


class CorpusLoader(Viewer):
    """
    Public interface for the CorpusLoader module. Maintains a reference to the logic Controller and the GUI wrapper.
    A CorpusLoader object can be used as a Panel component, i.e. will render in a Panel GUI.
    The callbacks added will be called when a corpus is built (can be set using set_build_callback()).
    """

    def __init__(self, root_directory: str, include_meta_loader: bool = False, **params):
        """
        :param root_directory: The root directory that the file selector will search for files to load. The argument must be a string. The directory may be non-existent at initialisation time, but no files will be displayed until it exists.
        :param include_meta_loader: If True, the Corpus Loader will include additional metadata joining functionality. False by default
        :param params: passed onto the panel.viewable.Viewer super-class
        :type root_directory: str
        :type include_meta_loader: bool
        """
        super().__init__(**params)
        self.controller: Controller = Controller(root_directory)
        self.view: ViewWrapperWidget = ViewWrapperWidget(self.controller, include_meta_loader)

    def __panel__(self):
        return self.view

    def add_tab(self, new_tab_name: str, new_tab_panel: Viewable):
        """
        Allows adding a Panel Viewable instance to the tab controls of the loader.
        :param new_tab_name: The name of the tab that will appear on the tab control bar
        :type new_tab_name: str
        :param new_tab_panel: The pane to attach to the new tab
        :type new_tab_panel: panel.viewable.Viewable
        """
        self.view.add_tab(new_tab_name, new_tab_panel)

    def register_event_callback(self, event_type: EventType, callback: Callable):
        """
        Registers a callback function to execute when the event specified by event_type occurs.
        Multiple callback functions can be registered and all will be called in order when the event occurs.
        When a callback raises an exception, the exception will be logged and the subsequent callbacks will be executed.
        The relevant corpus object will be passed as an argument for the BUILD and RENAME events.
        :param event_type: an enum with the possible values: LOAD, UNLOAD, BUILD, RENAME, DELETE
        :type event_type: EventType
        :param callback: the function to call when the event occurs
        :type callback: Callable
        """
        self.controller.register_event_callback(event_type, callback)

    def get_latest_corpus(self) -> Optional[DataFrameCorpus]:
        """
        :return: the last DataFrameCorpus object that was built. If none have been built, returns None.
        :rtype: Optional[DataFrameCorpus]
        """
        return self.controller.get_latest_corpus()

    def get_corpus(self, corpus_name: str) -> Optional[DataFrameCorpus]:
        """
        :return: the DataFrameCorpus corresponding to the provided name. If no corpus with the given name is found, return None
        :rtype: Optional[DataFrameCorpus]
        """
        return self.controller.get_corpus(corpus_name)

    def get_corpora(self) -> dict[str, DataFrameCorpus]:
        """
        :return: a dictionary that maps Corpus names to DataFrameCorpus objects that have been built using this CorpusLoader
        :rtype: dict[str, DataFrameCorpus]
        """
        return self.controller.get_corpora()

    def get_mutable_corpora(self) -> TCorpora:
        """
        Returns the corpora object that contains the loaded corpus objects.
        This allows adding to the corpora from outside the CorpusLoader as the object returned is mutable, not a copy.
        The Corpora object has a unique name constraint, meaning a corpus object cannot be added to the corpora if another
        corpus with the same name is already present. The same constraint applies to the rename method of corpus objects
        added to the corpora.
        :return: the mutable corpora object that contains the loaded corpus objects
        :rtype: TCorpora
        """
        return self.controller.get_mutable_corpora()



import logging
import traceback
from logging.handlers import RotatingFileHandler
from io import BytesIO
from os.path import abspath, join, dirname
from typing import Optional, Callable, Literal

import atap_corpus
from atap_corpus._types import TCorpora
from atap_corpus.corpus.corpus import DataFrameCorpus
from pandas import DataFrame
from panel.widgets import Tqdm

from atap_corpus_loader.controller.CorpusExportService import CorpusExportService
from atap_corpus_loader.controller.events import EventType, EventManager
from atap_corpus_loader.controller.loader_service import LoaderService
from atap_corpus_loader.controller.loader_service.FileLoadError import FileLoadError
from atap_corpus_loader.controller.loader_service.FileLoaderService import FileLoaderService
from atap_corpus_loader.controller.data_objects import (FileReference, ViewCorpusInfo, CorpusHeader, DataType,
                                                        UniqueNameCorpora)
from atap_corpus_loader.controller.loader_service.OniLoaderService import OniLoaderService
from atap_corpus_loader.controller.loader_service.file_loader_strategy.FileLoaderFactory import ValidFileType
from atap_corpus_loader.view.notifications import NotifierService


class Controller:
    LOGGER: logging.Logger = None
    """
    Provides methods for indirection between the corpus loading logic and the user interface
    Holds a reference to the latest corpus built.
    The callbacks will be called when a corpus is built (can be set using set_build_callback()).
    """

    def __init__(self, root_directory: str):
        Controller.setup_logger()

        self.file_loader_service: FileLoaderService = FileLoaderService(root_directory)
        self.oni_loader_service: OniLoaderService = OniLoaderService()
        self.loader_service: LoaderService = self.file_loader_service
        self.corpus_export_service: CorpusExportService = CorpusExportService()
        self.notifier_service: NotifierService = NotifierService()

        self.text_header: Optional[CorpusHeader] = None
        self.corpus_link_header: Optional[CorpusHeader] = None
        self.meta_link_header: Optional[CorpusHeader] = None

        self.corpus_headers: list[CorpusHeader] = []
        self.meta_headers: list[CorpusHeader] = []

        self.corpora: UniqueNameCorpora = UniqueNameCorpora()

        self.event_manager: EventManager = EventManager(self.LOGGER)

        self.build_tqdm = Tqdm(visible=False)
        self.export_tqdm = Tqdm(visible=False)

    @staticmethod
    def setup_logger():
        if Controller.LOGGER is not None:
            return
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        log_file_location = abspath(join(dirname(__file__), '..', 'log.txt'))
        # Max size is ~10MB with 1 backup, so a max size of ~20MB for log files
        max_bytes: int = 10000000
        backup_count: int = 1
        file_handler = RotatingFileHandler(log_file_location, maxBytes=max_bytes, backupCount=backup_count)
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(formatter)

        Controller.LOGGER = logging.getLogger(__name__)
        Controller.LOGGER.setLevel(logging.DEBUG)
        Controller.LOGGER.addHandler(file_handler)

        Controller.LOGGER.info('Logger started')

    def display_error(self, error_msg: str):
        Controller.LOGGER.error(f"Error displayed: {error_msg}")
        self.notifier_service.notify_error(error_msg)

    def display_success(self, success_msg: str):
        Controller.LOGGER.info(f"Success displayed: {success_msg}")
        self.notifier_service.notify_success(success_msg)

    def register_event_callback(self, event_type: EventType, callback: Callable):
        self.event_manager.register_event_callback(event_type, callback)

    def get_latest_corpus(self) -> Optional[DataFrameCorpus]:
        if len(self.corpora) == 0:
            return
        return self.corpora.items()[-1]

    def get_corpus(self, corpus_name: str) -> Optional[DataFrameCorpus]:
        return self.corpora.get(corpus_name)

    def get_corpora(self) -> dict[str, DataFrameCorpus]:
        corpora_list: list = self.corpora.items()
        corpora_dict: dict[str, DataFrameCorpus] = {}
        for corpus in corpora_list:
            corpora_dict[corpus.name] = corpus

        return corpora_dict

    def get_mutable_corpora(self) -> TCorpora:
        return self.corpora

    def set_loader_service_type(self, loader_type: Literal['file', 'oni']):
        if loader_type == 'file':
            self.loader_service = self.file_loader_service
        elif loader_type == 'oni':
            self.loader_service = self.oni_loader_service
        else:
            raise ValueError("loader_type specified must be either 'file' or 'oni'")

    def load_corpus_from_filepaths(self, filepath_ls: list[str], include_hidden: bool) -> bool:
        Controller.LOGGER.debug(f"Files loaded as corpus: {filepath_ls}")
        self.build_tqdm.visible = True
        try:
            self.loader_service.add_corpus_files(filepath_ls, include_hidden, self.build_tqdm)
            self.corpus_headers = self.loader_service.get_inferred_corpus_headers()
        except FileLoadError as e:
            self.LOGGER.error(traceback.format_exc())
            self.display_error(str(e))
            self.unload_all()
            self.build_tqdm.visible = False
            return False

        self.event_manager.trigger_callbacks(EventType.LOAD)
        self.build_tqdm.visible = False
        return True

    def load_meta_from_filepaths(self, filepath_ls: list[str], include_hidden: bool) -> bool:
        Controller.LOGGER.debug(f"Files loaded as meta: {filepath_ls}")
        self.build_tqdm.visible = True
        try:
            self.loader_service.add_meta_files(filepath_ls, include_hidden, self.build_tqdm)
            self.meta_headers = self.loader_service.get_inferred_meta_headers()
        except FileLoadError as e:
            self.LOGGER.error(traceback.format_exc())
            self.display_error(str(e))
            self.unload_all()
            self.build_tqdm.visible = False
            return False

        self.event_manager.trigger_callbacks(EventType.LOAD)
        self.build_tqdm.visible = False
        return True

    def build_corpus(self, corpus_id: str) -> bool:
        Controller.LOGGER.debug(f"build_corpus method: Building corpus with name: {corpus_id}")
        if self.is_meta_added():
            if (self.corpus_link_header is None) or (self.meta_link_header is None):
                self.display_error("Cannot build without link headers set. Select a corpus header and a meta header as linking headers in the dropdowns")
                return False

        self.build_tqdm.visible = True
        try:
            corpus = self.loader_service.build_corpus(corpus_id, self.corpus_headers,
                                                      self.meta_headers, self.text_header,
                                                      self.corpus_link_header, self.meta_link_header,
                                                      self.build_tqdm)
            Controller.LOGGER.debug(f"build_corpus method: corpus built")
        except FileLoadError as e:
            Controller.LOGGER.exception("Exception while building corpus: ")
            self.display_error(str(e))
            self.build_tqdm.visible = False
            return False
        except Exception as e:
            Controller.LOGGER.exception("Exception while building corpus: ")
            self.display_error(f"Unexpected error building corpus: {e}")
            self.build_tqdm.visible = False
            return False

        try:
            self.corpora.add(corpus)
            Controller.LOGGER.debug(f"build_corpus method: corpus added to corpora")
        except Exception as e:
            Controller.LOGGER.exception("Exception while adding corpus to corpora: ")
            self.display_error(str(e))
            self.build_tqdm.visible = False
            return False

        try:
            corpus.add_dtm(atap_corpus.parts.dtm.DTM.from_docs_with_vectoriser(corpus.docs()), 'tokens')
        except Exception as e:
            Controller.LOGGER.exception("Exception while building DTM: ")
            self.display_error(str(e))
            self.build_tqdm.visible = False
            return False

        self.event_manager.trigger_callbacks(EventType.BUILD, corpus)

        self.build_tqdm.visible = False
        Controller.LOGGER.debug(f"build_corpus method: corpus building complete")

        return True

    def get_corpora_info(self) -> list[ViewCorpusInfo]:
        corpora_info: list[ViewCorpusInfo] = []

        for corpus in reversed(self.corpora.items()):
            corpus_df: DataFrame = corpus.to_dataframe()

            parent_name: Optional[str] = None
            if corpus.parent is not None:
                parent_name = corpus.parent.name

            name: Optional[str] = corpus.name
            num_rows: int = len(corpus)
            headers: list[str] = []
            dtypes: list[str] = []
            dtype: str
            for header_name, dtype_obj in corpus_df.dtypes.items():
                try:
                    dtype = DataType(str(dtype_obj).lower()).name
                except ValueError:
                    dtype = DataType.TEXT.name
                dtypes.append(dtype)
                headers.append(str(header_name))
            first_row_data: list[str] = []
            if not corpus_df.empty:
                first_row_data = [str(x) for x in corpus_df.iloc[0]]

            corpora_info.append(ViewCorpusInfo(name, num_rows, parent_name, headers, dtypes, first_row_data))

        return corpora_info

    def delete_corpus(self, corpus_name: str):
        self.corpora.remove(corpus_name)
        self.event_manager.trigger_callbacks(EventType.DELETE)

    def rename_corpus(self, corpus_name: str, new_name: str):
        Controller.LOGGER.debug(f"Renaming corpus named '{corpus_name}' to '{new_name}'")
        corpus: Optional[DataFrameCorpus] = self.corpora.get(corpus_name)
        if corpus is None:
            self.display_error(f"No corpus with name {corpus_name} found")
            return

        try:
            corpus.rename(new_name)
        except ValueError as e:
            self.display_error(str(e))
        except Exception as e:
            self.display_error(f"Unexpected error while renaming: {e}")

        self.event_manager.trigger_callbacks(EventType.RENAME, corpus)

    def get_loaded_file_counts(self) -> dict[str, int]:
        corpus_file_set = self.loader_service.get_loaded_corpus_files()
        meta_file_set = self.loader_service.get_loaded_meta_files()
        file_set = corpus_file_set | meta_file_set

        file_counts: dict[str, int] = {"Total files": len(file_set)}
        for file_ref in file_set:
            extension = file_ref.get_extension().upper()
            if file_counts.get(extension) is None:
                file_counts[extension] = 1
            else:
                file_counts[extension] += 1

        return file_counts

    def unload_filepaths(self, filepath_ls: list[str]):
        for filepath in filepath_ls:
            self.loader_service.remove_meta_filepath(filepath)
            self.loader_service.remove_corpus_filepath(filepath)

        if not self.is_corpus_added():
            self.text_header = None
            self.corpus_headers = []
            self.corpus_link_header = None
        if not self.is_meta_added():
            self.meta_headers = []
            self.meta_link_header = None

        self.event_manager.trigger_callbacks(EventType.UNLOAD)

    def unload_all(self):
        Controller.LOGGER.debug("All files unloaded")
        self.loader_service.remove_all_files()

        self.text_header = None
        self.corpus_headers = []
        self.meta_headers = []
        self.corpus_link_header = None
        self.meta_link_header = None

        self.event_manager.trigger_callbacks(EventType.UNLOAD)

    def get_loaded_corpus_files(self) -> set[FileReference]:
        return self.loader_service.get_loaded_corpus_files()

    def get_loaded_meta_files(self) -> set[FileReference]:
        return self.loader_service.get_loaded_meta_files()

    def get_corpus_headers(self) -> list[CorpusHeader]:
        return self.corpus_headers

    def get_meta_headers(self) -> list[CorpusHeader]:
        return self.meta_headers

    def get_text_header(self) -> Optional[CorpusHeader]:
        return self.text_header

    def get_corpus_link_header(self) -> Optional[CorpusHeader]:
        return self.corpus_link_header

    def get_meta_link_header(self) -> Optional[CorpusHeader]:
        return self.meta_link_header

    def get_all_datatypes(self) -> list[str]:
        return [d.name for d in DataType]

    def get_valid_filetypes(self) -> list[str]:
        return [ft.name for ft in ValidFileType]

    def get_build_progress_bar(self) -> Tqdm:
        return self.build_tqdm

    def is_corpus_added(self) -> bool:
        return self.loader_service.is_corpus_loaded()

    def is_meta_added(self) -> bool:
        return self.loader_service.is_meta_loaded()

    def update_corpus_header(self, header: CorpusHeader, include: Optional[bool], datatype_name: Optional[str]):
        if include is not None:
            header.include = include
        if datatype_name is not None:
            header.datatype = DataType[datatype_name]

        for i, corpus_header in enumerate(self.corpus_headers):
            if header == corpus_header:
                self.corpus_headers[i] = header

    def update_meta_header(self, header: CorpusHeader, include: Optional[bool], datatype_name: Optional[str]):
        if include is not None:
            header.include = include
        if datatype_name is not None:
            header.datatype = DataType[datatype_name]

        for i, meta_header in enumerate(self.meta_headers):
            if header == meta_header:
                self.meta_headers[i] = header

    def set_text_header(self, text_header: Optional[str]):
        if text_header is None:
            self.text_header = None
            return

        for header in self.corpus_headers:
            if header.name == text_header:
                self.text_header = header
                header.datatype = DataType.TEXT
                header.include = True
                return

    def set_corpus_link_header(self, link_header_name: Optional[str]):
        for header in self.corpus_headers:
            if header.name == link_header_name:
                self.corpus_link_header = header
                header.include = True
                return
        self.corpus_link_header = None

    def set_meta_link_header(self, link_header_name: Optional[str]):
        for header in self.meta_headers:
            if header.name == link_header_name:
                self.meta_link_header = header
                header.include = True
                return
        self.meta_link_header = None

    def retrieve_all_files(self, expand_archived: bool) -> list[FileReference]:
        return self.loader_service.get_all_files(expand_archived)

    def get_export_types(self) -> list[str]:
        return self.corpus_export_service.get_filetypes()

    def export_corpus(self, corpus_name: str, filetype: str) -> Optional[BytesIO]:
        corpus: Optional[DataFrameCorpus] = self.corpora.get(corpus_name)
        if corpus is None:
            self.display_error(f"No corpus with name '{corpus_name}' found")
            return None

        self.export_tqdm.visible = True
        try:
            file_object: BytesIO = self.corpus_export_service.export(corpus, filetype, self.export_tqdm)
            self.export_tqdm.visible = False
            return file_object
        except ValueError as e:
            self.display_error(str(e))
        except Exception as e:
            self.display_error(f"Unexpected error while exporting: {e}")
        self.export_tqdm.visible = False

    def get_export_progress_bar(self) -> Tqdm:
        return self.export_tqdm

    # Oni Loader methods

    def set_provider(self, name: str, address: str) -> bool:
        try:
            self.oni_loader_service.set_provider(name, address)
        except FileLoadError as e:
            self.display_error(str(e))
            return False
        except Exception as e:
            self.display_error(f"Unexpected error while adding provider: {e}")
        return True

    def get_providers(self) -> list[str]:
        return self.oni_loader_service.get_providers()

    def set_curr_provider(self, name: str):
        try:
            self.oni_loader_service.set_curr_provider(name)
        except ValueError as e:
            self.display_error(str(e))
        except Exception as e:
            self.display_error(f"Unexpected error while setting provider: {e}")

    def get_curr_provider(self) -> str:
        return self.oni_loader_service.get_curr_provider()

    def get_curr_provider_address(self) -> str:
        return self.oni_loader_service.get_curr_provider_address()

    def set_api_key(self, api_key: str):
        success: bool = self.oni_loader_service.set_api_key(api_key)
        if success:
            self.display_success("API key set")
        else:
            self.display_error("API key not valid. Please try again")

    def set_collection_id(self, collection_id: str):
        try:
            self.oni_loader_service.set_collection_id(collection_id)
            self.unload_all()
            self.display_success(f"Collection files retrieved successfully for '{collection_id}'")
        except FileLoadError as e:
            self.display_error(str(e))
        except Exception as e:
            self.display_error(f"Unexpected error while setting collection ID: {e}")



import panel


class NotifierService:
    """
    Provides an object to handle GUI notifications as an indirection from the Panel library notifications.
    """
    def __init__(self):
        panel.extension(notifications=True)

    def notify_error(self, error_msg: str):
        """
        Renders the provided error message. The notification will last indefinitely until the user dismisses it.
        :param error_msg: The error message to be displayed as a string
        """
        panel.state.notifications.error(error_msg, duration=0)

    def notify_success(self, success_msg: str):
        """
        Renders the provided success message. The notification will last for 3 seconds or until the user dismisses it.
        :param success_msg: The success message to be displayed as a string
        """
        panel.state.notifications.success(success_msg, duration=3000)



import traceback
from typing import Callable, Optional

from atap_corpus_loader.controller.events import EventType


class EventManager:
    def __init__(self, logger):
        self.logger = logger

        self.callback_mapping: dict[EventType, list[Callable]] = {}
        self.reset_callbacks()

    def reset_callbacks(self):
        self.callback_mapping = {e: [] for e in EventType}

    def register_event_callback(self, event_type: EventType, callback: Callable):
        if not callable(callback):
            raise TypeError("Provided callback function must be callable")
        self.callback_mapping[event_type].append(callback)
        self.logger.info(f"New callback registered for event '{event_type.name}'. Callback: {callback}")

    def trigger_callbacks(self, event_type: EventType, *callback_args):
        callback_list: Optional[list[Callable]] = self.callback_mapping.get(event_type)
        if callback_list is None:
            raise ValueError(f"No callbacks registered for event type: {event_type.name}")
        for callback in callback_list:
            try:
                callback(*callback_args)
                self.logger.info(f"Callback executed for event '{event_type.name}'. Callback: {callback}")
            except Exception as e:
                self.logger.exception(f"Exception while executing callback for event '{event_type.name}': {type(e)}\n{traceback.format_exc()}")



import logging
from typing import Optional, Iterable

from atap_corpus._types import TCorpus
from atap_corpus.corpus.base import BaseCorpora, BaseCorpus
from atap_corpus.utils import format_dunder_str

logger = logging.getLogger(__name__)


class UniqueNameCorpora(BaseCorpora):
    """
    UniqueNameCorpora is a container for BaseCorpus objects that ensures all corpus objects have unique names within the Corpora.
    Additionally, while stored within the Corpora, the rename method of the BaseCorpus objects is replaced with a rename method
    that ensures the new name is unique.
    """

    def __init__(self, corpus: Optional[BaseCorpus | Iterable[BaseCorpus]] = None):
        super().__init__(corpus)
        self._collection = dict()
        if corpus is None:
            return
        elif isinstance(corpus, Iterable):
            for c in corpus:
                if c.name in self._collection.keys():
                    logger.warning(f"Corpus name: {c.name} is duplicated. Only one is kept for uniqueness.")
                else:
                    self.add(c)
        elif isinstance(corpus, BaseCorpus):
            self.add(corpus)
        else:
            raise TypeError(f"Corpora can only store Corpus objects. Got {corpus.__class__.__name__}.")

    def _verify_unique_name(self, name: str):
        if name in self._collection:
            raise ValueError(f"Corpus with name '{name}' already exists. Select a different name")

    def _unique_rename(self, corpus: TCorpus, name: str):
        self._verify_unique_name(name)
        old_name = corpus.name
        corpus._name = name
        self._collection[name] = self._collection.pop(old_name)

    def add(self, corpus: TCorpus):
        """ Adds a Corpus into the Corpora. Corpus name is used as the name for get(), remove().
        If a corpus with the same name is added again, a ValueError is raised.
        If a corpus is added with a name that is empty or None, it is renamed to a generated name based on the current time.
        BaseCorpus objects added have their rename method replaced to include a check for name uniqueness.
        """
        if not isinstance(corpus, BaseCorpus):
            raise TypeError(f"Corpora can only store Corpus objects. Got {corpus.__class__.__name__}.")

        self._verify_unique_name(corpus.name)

        corpus.__ORIG_RENAME = corpus.rename
        corpus.rename = lambda new_name, corpus_obj=corpus: self._unique_rename(corpus_obj, new_name)
        self._collection[corpus.name] = corpus

    def remove(self, name: str):
        """ Remove a Corpus from the Corpora.
        The rename method of the Corpus object has the unique name constraint removed before it is removed from the Corpora.
        If Corpus does not exist, it'll have no effect.
        """
        try:
            self._collection[name].rename = self._collection[name].__ORIG_RENAME
            del self._collection[name]
        except KeyError:
            return

    def items(self) -> list[TCorpus]:
        """
        Returns a shallow copy list of the corpus objects in the Corpora, in order of addition.
        :return: A list of corpus objects found in the Corpora
        :rtype: list[TCorpus]
        """
        return list(self._collection.values())

    def get(self, name: str) -> Optional[TCorpus]:
        """
        Returns the corpus object corresponding with the provided name.
        If no corpus exists with the name, returns None.
        :param name: name of the corpus to be retrieved
        :type name: str
        :return: The corresponding corpus, or None if not found
        :rtype: Optional[TCorpus]
        """
        return self._collection.get(name)

    def clear(self):
        """
        Clears the Corpora of all corpus objects.
        """
        self._collection = dict()

    def __len__(self) -> int:
        """
        :return: the number of Corpus in the Corpora.
        :rtype: int
        """
        return len(self._collection)

    def __str__(self) -> str:
        return format_dunder_str(self.__class__, **{"size": len(self)})



from io import BytesIO, StringIO
from typing import Callable
from zipfile import ZipFile

import numpy as np
from atap_corpus.corpus.corpus import DataFrameCorpus
from pandas import DataFrame, ExcelWriter
from panel.widgets import Tqdm


class CorpusExportService:
    def __init__(self):
        self.export_type_mapping: dict[str, Callable] = {
            'csv': self.export_csv,
            'xlsx': self.export_xlsx,
            'zip': self.export_zip
        }

    def get_filetypes(self) -> list[str]:
        return list(self.export_type_mapping.keys())

    def export(self, corpus: DataFrameCorpus, filetype: str, tqdm_obj: Tqdm) -> BytesIO:
        if filetype not in self.export_type_mapping:
            raise ValueError(f"{filetype} is not a valid export format")
        file_object: BytesIO = self.export_type_mapping[filetype](corpus, tqdm_obj)
        file_object.seek(0)

        return file_object

    @staticmethod
    def export_csv(corpus: DataFrameCorpus, tqdm_obj: Tqdm) -> BytesIO:
        csv_object = BytesIO()
        if len(corpus) == 0:
            return csv_object

        df: DataFrame = corpus.to_dataframe()
        chunks = np.array_split(df.index, min(len(df), 1000))
        with tqdm_obj(total=len(df), desc="Exporting to CSV", unit="documents", leave=False) as pbar:
            df.loc[chunks[0]].to_csv(csv_object, mode='w', index=False)
            pbar.update(len(chunks[0]))
            for chunk, subset in enumerate(chunks[1:]):
                df.loc[subset].to_csv(csv_object, header=False, mode='a', index=False)
                pbar.update(len(subset))

        return csv_object

    @staticmethod
    def export_xlsx(corpus: DataFrameCorpus, tqdm_obj: Tqdm) -> BytesIO:
        excel_object = BytesIO()
        if len(corpus) == 0:
            return excel_object

        df: DataFrame = corpus.to_dataframe()
        chunks = np.array_split(df.index, min(len(df), 1000))
        with tqdm_obj(total=len(df), desc="Exporting to Excel", unit="documents", leave=False) as pbar:
            with ExcelWriter(excel_object) as writer:
                df.loc[chunks[0]].to_excel(writer, index=False, header=True, sheet_name='Sheet1')
                pbar.update(len(chunks[0]))
                for chunk, subset in enumerate(chunks[1:]):
                    df.loc[subset].to_excel(writer, startrow=subset[0]+1, index=False, header=False, sheet_name='Sheet1')
                    pbar.update(len(subset))

        return excel_object

    @staticmethod
    def export_zip(corpus: DataFrameCorpus, tqdm_obj: Tqdm) -> BytesIO:
        zipped_object = BytesIO()
        if len(corpus) == 0:
            return zipped_object

        df: DataFrame = corpus.to_dataframe()
        metas_df: DataFrame = df[corpus.metas]
        filename_col = 'filename'
        while filename_col in metas_df.columns:
            filename_col = filename_col + '_'
        metas_df.rename({'filename': filename_col})
        metas_df['filename'] = ''

        zip_file = ZipFile(zipped_object, mode='w')
        for i in tqdm_obj(range(len(corpus)), desc="Exporting to zipped file", unit="documents", leave=False):
            sanitised_name = "".join(c for c in corpus.name if c.isalpha() or c.isdigit() or c == ' ').rstrip()
            filename = f"{sanitised_name}-{i + 1}.txt"
            document = str(corpus[i])
            zip_file.writestr(filename, document)
            metas_df.loc[i, 'filename'] = filename

        if len(corpus.metas) > 0:
            metadata_buffer = StringIO()
            metas_df.to_csv(metadata_buffer, index=False, mode="w")
            zip_file.writestr('metadata.csv', metadata_buffer.getvalue())
            zip_file.close()

        return zipped_object



from io import BytesIO
from typing import Optional

import requests
from panel.widgets import Tqdm

from atap_corpus_loader.controller.data_objects import FileReference
from atap_corpus_loader.controller.data_objects.FileReference import RemoteFileReference
from atap_corpus_loader.controller.loader_service import LoaderService
from atap_corpus_loader.controller.loader_service.FileLoadError import FileLoadError


class OniLoaderService(LoaderService):
    def __init__(self):
        super().__init__()
        self.providers: dict[str, str] = {
            "LDaCA": "https://data.ldaca.edu.au",
            "ATAP": "https://data.atap.edu.au"
        }
        self.curr_provider: str = "LDaCA"
        self.api_key: Optional[str] = None
        self.collection_id: str = ''
        self.collection_files: list[FileReference] = []

    def _get_api_root(self) -> str:
        base_url: Optional[str] = self.providers.get(self.curr_provider)
        if base_url is None:
            return ''
        return base_url + '/api/'

    def _get_auth_header(self) -> dict:
        return {'Authorization': f"Bearer {self.api_key}"}

    def _validate_collection_id(self, collection_id: str) -> bool:
        if len(collection_id) == 0:
            return False

        return True

    def _validate_api_key(self, api_key: str) -> bool:
        if len(api_key) == 0:
            return False

        return True

    def set_provider(self, name: str, address: str):
        if name == '':
            raise ValueError("The name cannot be empty")
        if address == '':
            raise ValueError("The address cannot be empty")
        self.providers[name] = address

    def get_providers(self) -> list[str]:
        return list(self.providers.keys())

    def set_curr_provider(self, name: str):
        if name not in self.providers:
            raise ValueError(f"Provider '{name}' not found in providers list")
        self.curr_provider = name

    def get_curr_provider(self) -> str:
        return self.curr_provider

    def get_curr_provider_address(self) -> str:
        curr_address = self.providers.get(self.curr_provider)
        if curr_address:
            return curr_address
        return ''

    def set_api_key(self, api_key: str) -> bool:
        if self._validate_api_key(api_key):
            self.api_key = api_key
            return True
        return False

    def set_collection_id(self, collection_id: str):
        collection_id = collection_id.strip()
        if not self._validate_collection_id(collection_id):
            raise FileLoadError("Collection ID is malformed")
        self.collection_id = collection_id
        self.retrieve_collection_files()

    def retrieve_collection_files(self):
        if self.api_key is None:
            self.collection_files = []
            raise FileLoadError("No API key set")

        api_root: str = self._get_api_root()

        r = requests.get(
            api_root + "object/meta",
            params={"id": self.collection_id, "noUrid": "1"},
        )

        try:
            r.raise_for_status()
        except Exception as e:
            self.collection_files = []
            raise FileLoadError(str(e))

        metadata = r.json()
        if metadata.get('error') is not None:
            self.collection_files = []
            raise FileLoadError('Error retrieving collection. Ensure the collection ID is correct')
        metadata_items = metadata.get('@graph')
        if metadata_items is None:
            self.collection_files = []
            raise FileLoadError('No items found in collection')

        all_file_refs: list[FileReference] = []
        for item in metadata_items:
            item_type = item.get("@type")
            if item_type == "RepositoryObject":
                filepath = item["indexableText"]["@id"]

                file_refs: list[FileReference] = self.file_ref_factory.get_file_refs_from_path(filepath, False)
                all_file_refs.extend(file_refs)

        all_file_refs.sort(key=lambda ref: ref.get_path())

        self.collection_files = all_file_refs

    def get_all_files(self, expand_archived: bool) -> list[FileReference]:
        return self.collection_files

    def add_corpus_files(self, corpus_filepaths: list[str], include_hidden: bool, tqdm_obj: Tqdm):
        for filepath in tqdm_obj(corpus_filepaths, desc="Retrieving corpus files", unit="files", leave=False):
            file_ref: RemoteFileReference = self.file_ref_factory.get_oni_file_ref(filepath)
            if file_ref in self.loaded_corpus_files:
                continue
            if not include_hidden and file_ref.is_hidden():
                continue
            data = requests.get(
                self._get_api_root() + "stream",
                params={
                    "id": self.collection_id,
                    "path": filepath,
                },
                headers=self._get_auth_header()
            )

            try:
                data.raise_for_status()
            except Exception as e:
                if str(e).startswith('404'):
                    continue
                elif str(e).startswith('401'):
                    raise FileLoadError(f"Denied permission to access file: {filepath}<br>Check API Key")
                else:
                    raise FileLoadError(f"Unexpected error loading file: {filepath}<br>{str(e)}")

            content_buf = BytesIO(data.text.encode('utf-8'))
            file_ref.set_content_buffer(content_buf)

            self.loaded_corpus_files.add(file_ref)

    def add_meta_files(self, meta_filepaths: list[str], include_hidden: bool, tqdm_obj: Tqdm):
        for filepath in tqdm_obj(meta_filepaths, desc="Retrieving metadata files", unit="files", leave=False):
            file_ref: RemoteFileReference = self.file_ref_factory.get_oni_file_ref(filepath)
            if file_ref in self.loaded_meta_files:
                continue
            if not include_hidden and file_ref.is_hidden():
                continue
            data = requests.get(
                self._get_api_root() + "stream",
                params={
                    "id": self.collection_id,
                    "path": filepath,
                },
                headers=self._get_auth_header()
            )

            try:
                data.raise_for_status()
            except Exception as e:
                if str(e).startswith('404'):
                    continue
                elif str(e).startswith('401'):
                    raise FileLoadError(f"Denied permission to access file: {filepath}<br>Check API Key")
                else:
                    raise FileLoadError(f"Unexpected error loading file: {filepath}<br>{str(e)}")

            content_buf = BytesIO()
            content_buf.write(data.text.encode('utf-8'))
            file_ref.set_content_buffer(content_buf)

            self.loaded_meta_files.add(file_ref)



from typing import Optional

from panel import Tabs
from panel.viewable import Viewable
from panel.widgets import TooltipIcon

from atap_corpus_loader.controller import Controller
from atap_corpus_loader.view.gui import AbstractWidget, FileLoaderWidget, CorpusInfoWidget, OniLoaderWidget
from atap_corpus_loader.view.tooltips import TooltipManager


class ViewWrapperWidget(AbstractWidget):
    """
    A wrapper class that holds different loading method interfaces within a Tab
    """
    def __init__(self, controller: Controller, include_meta_loader: bool):
        super().__init__()
        self.controller: Controller = controller
        self.tooltip_manager: TooltipManager = TooltipManager()

        self.file_loader: FileLoaderWidget = FileLoaderWidget(self, controller, include_meta_loader)
        self.oni_loader: OniLoaderWidget = OniLoaderWidget(self, controller, include_meta_loader)
        self.corpus_display: CorpusInfoWidget = CorpusInfoWidget(controller)

        # set_load_service_type depends on the order of these tabs
        self.panel = Tabs(("File Loader", self.file_loader),
                          ("Oni Loader", self.oni_loader),
                          ("Corpus Overview", self.corpus_display),
                          dynamic=True)
        self.panel.param.watch(self.set_load_service_type, parameter_names=['active'])
        self.corpus_info_idx: int = len(self.panel) - 1
        self.children = [self.file_loader, self.oni_loader, self.corpus_display]

    def update_display(self):
        pass

    def add_tab(self, new_tab_name: str, new_tab_panel: Viewable):
        self.panel.append((new_tab_name, new_tab_panel))

    def load_corpus_from_filepaths(self, filepath_ls: list[str], include_hidden: bool):
        if len(filepath_ls) == 0:
            return
        success = self.controller.load_corpus_from_filepaths(filepath_ls, include_hidden)
        self.update_displays()
        if success:
            self.controller.display_success("Corpus files loaded successfully")

    def load_meta_from_filepaths(self, filepath_ls: list[str], include_hidden: bool):
        if len(filepath_ls) == 0:
            return
        success = self.controller.load_meta_from_filepaths(filepath_ls, include_hidden)
        self.update_displays()
        if success:
            self.controller.display_success("Metadata files loaded successfully")

    def build_corpus(self, corpus_id: str) -> bool:
        success: bool = self.controller.build_corpus(corpus_id)
        if success:
            self.update_displays()

            self.panel.active = self.corpus_info_idx
            corpus_id: str = self.controller.get_latest_corpus().name
            self.controller.display_success(f"Corpus {corpus_id} built successfully")

        return success

    def set_load_service_type(self, *_):
        active_tab: int = self.panel.active
        if active_tab == 0:
            self.controller.set_loader_service_type('file')
        elif active_tab == 1:
            self.controller.set_loader_service_type('oni')
        self.file_loader.unload_all()

    def get_tooltip(self, tooltip_name: str) -> Optional[TooltipIcon]:
        return self.tooltip_manager.get_tooltip(tooltip_name)



from abc import ABC, abstractmethod

from pandas import DataFrame

from atap_corpus_loader.controller.data_objects import CorpusHeader, FileReference


class FileLoaderStrategy(ABC):
    """
    An abstract class for loading files as DataFrame objects to be used in a DataFrameCorpus.
    A concrete class should extend this class for each file type that is supported.
    """
    def __init__(self, file_ref: FileReference):
        """
        :param file_ref: the FileReference object corresponding to the file to be loaded
        """
        self.file_ref: FileReference = file_ref

    @staticmethod
    def _apply_selected_dtypes(df: DataFrame, headers: list[CorpusHeader]) -> DataFrame:
        """
        Attempts to cast each column within the provided DataFrame to the data types specified in headers.
        :param df: the DataFrame object whose columns will be type cast
        :param headers: the CorpusHeader objects representing the columns of the DataFrame. CorpusHeader objects with
        include as False will be ignored.
        :return: the DataFrame with columns cast to the given data types
        """
        dtypes = {h.name: h.datatype.value for h in headers if h.include}

        return df.astype(dtype=dtypes)

    @abstractmethod
    def get_inferred_headers(self) -> list[CorpusHeader]:
        """
        Provides a list of CorpusHeader objects corresponding to the data found within the file.
        Some additional metadata headers may be provided not found within the file, such as filepath
        :return: a list of CorpusHeader objects corresponding to the data found within the file
        """
        raise NotImplementedError()

    @abstractmethod
    def get_dataframe(self, headers: list[CorpusHeader]) -> DataFrame:
        """
        Provides a DataFrame object containing the data from the loaded file.
        Columns of the DataFrame will be cast to the data types specified in the headers parameter.
        The DataFrame will exclude a column of data if its corresponding CorpusHeader object has include set to False
        :param headers: a list of CorpusHeader objects corresponding to the data found within the file
        :return: a DataFrame object corresponding to the loaded file and its provided CorpusHeader list
        """
        raise NotImplementedError()



from io import BytesIO

from pandas import DataFrame, read_excel

from atap_corpus_loader.controller.data_objects import CorpusHeader, DataType
from atap_corpus_loader.controller.loader_service.file_loader_strategy.FileLoaderStrategy import FileLoaderStrategy


class XLSXLoaderStrategy(FileLoaderStrategy):
    def get_inferred_headers(self) -> list[CorpusHeader]:
        file_buf: BytesIO = self.file_ref.get_content_buffer()
        df: DataFrame = read_excel(file_buf, header=0, nrows=2)
        headers: list[CorpusHeader] = []
        for header_name, dtype_obj in df.dtypes.items():
            dtype: DataType
            try:
                dtype = DataType(str(dtype_obj))
            except ValueError:
                dtype = DataType.TEXT
            headers.append(CorpusHeader(str(header_name), dtype))

        return headers

    def get_dataframe(self, headers: list[CorpusHeader]) -> DataFrame:
        included_headers: list[str] = [header.name for header in headers if header.include]
        file_buf: BytesIO = self.file_ref.get_content_buffer()
        df: DataFrame = read_excel(file_buf, dtype=object, header=0, usecols=included_headers)
        dtypes_applied_df: DataFrame = FileLoaderStrategy._apply_selected_dtypes(df, headers)

        return dtypes_applied_df



from panel import Row, Column
from panel.layout import Divider
from panel.pane import Markdown
from panel.widgets import Select, TextInput, Button, PasswordInput

from atap_corpus_loader.controller import Controller
from atap_corpus_loader.view import ViewWrapperWidget
from atap_corpus_loader.view.gui import AbstractWidget, FileLoaderWidget


class OniLoaderWidget(AbstractWidget):
    def __init__(self, view_handler: ViewWrapperWidget, controller: Controller, include_meta_loader: bool):
        super().__init__()
        self.view_handler: ViewWrapperWidget = view_handler
        self.controller: Controller = controller

        provider_tooltip = self.view_handler.get_tooltip('oni_provider')
        self.provider_selector = Select(name='Provider selector', width=150)
        self.provider_selector.param.watch(self._on_provider_change, ['value'])

        self.toggle_add_provider_button = Button(name='Add new provider', width=150, button_type='primary', button_style='outline', align='center')
        self.toggle_add_provider_button.on_click(self._toggle_show_add_provider_pane)

        self.add_provider_name_input = TextInput(name='Provider name', placeholder='Provider name, e.g. MyProvider', visible=False)
        self.add_provider_address_input = TextInput(name='Provider address', placeholder='Provider address, e.g. https://data.ldaca.edu.au', visible=False)
        self.add_provider_button = Button(name='Add', button_type='success', button_style='solid', visible=False)
        self.add_provider_button.on_click(self._add_provider)
        self.add_provider_panel: Column = Column(self.add_provider_name_input,
                                                 self.add_provider_address_input,
                                                 self.add_provider_button,
                                                 visible=False)

        api_key_tooltip = self.view_handler.get_tooltip('oni_api_key')
        self.api_key_input = PasswordInput(name='API Key', placeholder='af6391e0-f873-11ee-8355-bae397411a92')
        self.api_key_input.param.watch(self._set_api_key, ['value'])

        retrieve_collection_tooltip = self.view_handler.get_tooltip('oni_retrieve_collection')
        self.collection_id_input = TextInput(name='Collection ID', placeholder='arcp://name,doi10.26180%2F23961609')
        self.retrieve_collection_button = Button(name='Retrieve collection information', button_type='success', button_style='solid', align='end')
        self.retrieve_collection_button.on_click(self._retrieve_collection_information)

        self.file_loader: FileLoaderWidget = FileLoaderWidget(view_handler, controller, include_meta_loader)

        self.curr_provider_display = Markdown(stylesheets=["a { margin: 0;}"])

        self.panel = Column(
            Row(self.provider_selector,
                self.curr_provider_display,
                self.toggle_add_provider_button,
                provider_tooltip),
            Row(self.add_provider_panel),
            Row(self.api_key_input,
                api_key_tooltip),
            Row(self.collection_id_input,
                self.retrieve_collection_button,
                retrieve_collection_tooltip),
            Divider(),
            Row(self.file_loader),
            sizing_mode='stretch_width'
        )
        self.children = [self.file_loader]
        self.update_display()

    def update_display(self):
        self.provider_selector.options = self.controller.get_providers()
        self.provider_selector.value = self.controller.get_curr_provider()
        curr_provider_address = self.controller.get_curr_provider_address()
        link = f'<a href="{curr_provider_address}" target="_blank">{curr_provider_address}</a>'
        self.curr_provider_display.object = link

    def _on_provider_change(self, *_):
        self.controller.set_curr_provider(self.provider_selector.value)
        curr_provider_address = self.controller.get_curr_provider_address()
        link = f'<a href="{curr_provider_address}" target="_blank">{curr_provider_address}</a>'
        self.curr_provider_display.object = link

    def _toggle_show_add_provider_pane(self, *_):
        add_provider_style = self.toggle_add_provider_button.button_style
        if add_provider_style == 'outline':
            self.toggle_add_provider_button.button_style = 'solid'
        elif add_provider_style == 'solid':
            self.toggle_add_provider_button.button_style = 'outline'

        self.add_provider_panel.visible = not self.add_provider_panel.visible
        self.add_provider_name_input.visible = not self.add_provider_name_input.visible
        self.add_provider_address_input.visible = not self.add_provider_address_input.visible
        self.add_provider_button.visible = not self.add_provider_button.visible

    def _add_provider(self, *_):
        provider_name: str = self.add_provider_name_input.value
        provider_address: str = self.add_provider_address_input.value
        self.controller.set_provider(provider_name, provider_address)
        self.add_provider_name_input.value = ''
        self.add_provider_address_input.value = ''
        self._toggle_show_add_provider_pane()
        self.provider_selector.value = provider_name

        self.update_display()

    def _set_api_key(self, *_):
        api_key: str = self.api_key_input.value
        self.controller.set_api_key(api_key)

    def _retrieve_collection_information(self, *_):
        collection_id: str = self.collection_id_input.value
        self.controller.set_collection_id(collection_id)
        self.file_loader.unload_all()
        self.view_handler.update_displays()



from typing import Optional

from panel import Column, GridBox, bind, Row, Spacer
from panel.pane import Markdown, Str
from panel.widgets import Select, Checkbox, Button

from atap_corpus_loader.controller import Controller
from atap_corpus_loader.controller.data_objects.CorpusHeader import CorpusHeader
from atap_corpus_loader.view import ViewWrapperWidget
from atap_corpus_loader.view.gui import AbstractWidget


class MetaEditorWidget(AbstractWidget):
    TABLE_BORDER_STYLE = {'border': '1px dashed black', 'border-radius': '5px'}
    ERROR_BORDER_STYLE = {'border': '1px solid red', 'border-radius': '5px'}
    HEADER_STYLE = {'margin-top': '0', 'margin-bottom': '0'}

    def __init__(self, view_handler: ViewWrapperWidget, controller: Controller):
        super().__init__()
        self.view_handler: ViewWrapperWidget = view_handler
        self.controller: Controller = controller

        self.corpus_table_container = GridBox(styles=MetaEditorWidget.TABLE_BORDER_STYLE)
        self.meta_table_container = GridBox(styles=MetaEditorWidget.TABLE_BORDER_STYLE)

        corpus_table_title = Markdown("## Corpus editor")
        corpus_table_tooltip = self.view_handler.get_tooltip('corpus_editor')
        self.corpus_table_row: Row = Row(corpus_table_tooltip, corpus_table_title)

        self.text_header_dropdown = Select(name='Select document label', width=200)
        self.corpus_checkboxes: list[Checkbox] = []
        self.corpus_include_all_Button = Button(name="Include all", button_style='outline', button_type='primary', align='end')
        self.corpus_include_all_Button.on_click(self._include_all_corpus)

        meta_table_title = Markdown("## Metadata editor")
        meta_table_tooltip = self.view_handler.get_tooltip('meta_editor')
        self.meta_table_row: Row = Row(meta_table_tooltip, meta_table_title)
        self.meta_checkboxes: list[Checkbox] = []
        self.meta_include_all_button = Button(name="Include all", button_style='outline', button_type='primary')
        self.meta_include_all_button.on_click(self._include_all_meta)

        self.link_row = Row(visible=False, styles=MetaEditorWidget.ERROR_BORDER_STYLE)
        link_row_tooltip = self.view_handler.get_tooltip('linking_selectors')
        self.corpus_link_dropdown = Select(name='Select corpus linking label', width=180)
        self.meta_link_dropdown = Select(name='Select metadata linking label', width=180)
        link_emoji = '\U0001F517'
        self.link_markdown = Str(link_emoji, styles={"font-size": "2em", "margin": "auto"})
        self.link_row.objects = [link_row_tooltip,
                                 self.corpus_link_dropdown,
                                 self.link_markdown.clone(),
                                 self.meta_link_dropdown]

        self.corpus_editor_control_row = Row(self.text_header_dropdown, self.corpus_include_all_Button, visible=False)
        self.meta_editor_control_row = Row(self.meta_include_all_button, visible=False)

        text_header_fn = bind(self._set_text_header, self.text_header_dropdown)
        corpus_link_fn = bind(self._set_corpus_link_header, self.corpus_link_dropdown)
        meta_link_fn = bind(self._set_meta_link_header, self.meta_link_dropdown)
        self.panel = Column(
            self.corpus_table_row,
            self.corpus_editor_control_row,
            self.corpus_table_container,
            Spacer(height=20),
            self.link_row,
            Spacer(height=20),
            self.meta_table_row,
            self.meta_editor_control_row,
            self.meta_table_container,
            text_header_fn, corpus_link_fn, meta_link_fn
        )
        self.update_display()

    def update_display(self):
        self._build_corpus_table()
        self._build_meta_table()
        self._update_dropdowns()

    def _include_all_corpus(self, *_):
        for checkbox in self.corpus_checkboxes:
            checkbox.value = True

    def _include_all_meta(self, *_):
        for checkbox in self.meta_checkboxes:
            checkbox.value = True

    def _set_text_header(self, text_header_name: Optional[str]):
        self.controller.set_text_header(text_header_name)
        self.update_display()

    def _set_corpus_link_header(self, header_name: str):
        self.controller.set_corpus_link_header(header_name)
        self.update_display()

    def _set_meta_link_header(self, header_name: str):
        self.controller.set_meta_link_header(header_name)
        self.update_display()

    def _get_table_cells_list(self, headers: list[CorpusHeader], link_header: CorpusHeader, is_meta_table: bool) -> tuple[int, list]:
        all_datatypes: list[str] = self.controller.get_all_datatypes()
        text_header: Optional[CorpusHeader] = self.controller.get_text_header()

        if is_meta_table:
            self.meta_checkboxes = []
        else:
            self.corpus_checkboxes = []

        table_cells: list = [Markdown('**Data label**', align='start'),
                             Markdown('**Datatype**', align='start'),
                             Markdown('**Include**', align='center')]
        if self.controller.is_meta_added():
            table_cells.append(Markdown('**Link**', align='center'))
        ncols: int = len(table_cells)

        for i, header in enumerate(headers):
            is_text = (header == text_header) and (not is_meta_table)
            is_link = (header == link_header)

            if is_link:
                header.include = True

            table_cells.append(Markdown(header.name, align='start', styles=MetaEditorWidget.HEADER_STYLE))

            datatype_selector = Select(options=all_datatypes, value=header.datatype.name, width=120, disabled=is_text)
            if is_meta_table:
                dtype_fn = bind(self.controller.update_meta_header, header, None, datatype_selector)
            else:
                dtype_fn = bind(self.controller.update_corpus_header, header, None, datatype_selector)
            table_cells.append(Row(datatype_selector, Column(dtype_fn, visible=False)))

            include_checkbox = Checkbox(value=header.include, align='center', disabled=(is_text or is_link))
            if is_meta_table:
                self.meta_checkboxes.append(include_checkbox)
                include_fn = bind(self.controller.update_meta_header, header, include_checkbox, None)
            else:
                self.corpus_checkboxes.append(include_checkbox)
                include_fn = bind(self.controller.update_corpus_header, header, include_checkbox, None)
            table_cells.append(Row(include_checkbox, Column(include_fn, visible=False)))

            if self.controller.is_meta_added():
                if is_link:
                    link_identifier = self.link_markdown.clone()
                else:
                    link_identifier = ' '
                table_cells.append(link_identifier)

        return ncols, table_cells

    def _build_corpus_table(self):
        is_corpus_added = self.controller.is_corpus_added()
        self.corpus_table_row.visible = is_corpus_added
        self.corpus_editor_control_row.visible = is_corpus_added
        self.corpus_table_container.visible = is_corpus_added

        corpus_headers: list[CorpusHeader] = self.controller.get_corpus_headers()
        link_header: Optional[CorpusHeader] = self.controller.get_corpus_link_header()

        ncols, corpus_table_cells = self._get_table_cells_list(corpus_headers, link_header, False)

        self.corpus_table_container.objects = corpus_table_cells
        self.corpus_table_container.ncols = ncols

    def _build_meta_table(self):
        is_meta_added = self.controller.is_meta_added()
        self.meta_table_row.visible = is_meta_added
        self.meta_editor_control_row.visible = is_meta_added
        self.meta_table_container.visible = is_meta_added

        meta_headers: list[CorpusHeader] = self.controller.get_meta_headers()
        link_header: Optional[CorpusHeader] = self.controller.get_meta_link_header()

        ncols, meta_table_cells = self._get_table_cells_list(meta_headers, link_header, True)

        self.meta_table_container.objects = meta_table_cells
        self.meta_table_container.ncols = ncols

    def _update_dropdowns(self):
        is_meta_added = self.controller.is_meta_added()
        is_corpus_added = self.controller.is_corpus_added()
        self.link_row.visible = is_meta_added
        self.text_header_dropdown.visible = is_corpus_added

        corpus_headers: list[CorpusHeader] = self.controller.get_corpus_headers()
        meta_headers: list[CorpusHeader] = self.controller.get_meta_headers()
        text_header: Optional[CorpusHeader] = self.controller.get_text_header()
        corpus_link_header: Optional[CorpusHeader] = self.controller.get_corpus_link_header()
        meta_link_header: Optional[CorpusHeader] = self.controller.get_meta_link_header()

        self.text_header_dropdown.options = [h.name for h in corpus_headers]
        if text_header is not None:
            self.text_header_dropdown.value = text_header.name

        self.corpus_link_dropdown.options = [''] + [h.name for h in corpus_headers]
        if corpus_link_header is None:
            self.corpus_link_dropdown.value = ''
        else:
            self.corpus_link_dropdown.value = corpus_link_header.name

        self.meta_link_dropdown.options = [''] + [h.name for h in meta_headers]
        if meta_link_header is None:
            self.meta_link_dropdown.value = ''
        else:
            self.meta_link_dropdown.value = meta_link_header.name

        if (meta_link_header is None) or (corpus_link_header is None):
            self.link_row.styles = MetaEditorWidget.ERROR_BORDER_STYLE
        else:
            self.link_row.styles = {}



from glob import iglob
from os import R_OK, access
from os.path import normpath, sep, isdir, exists
from typing import Iterator
from zipfile import BadZipFile

from panel.widgets import Tqdm

from atap_corpus_loader.controller.data_objects import FileReference
from atap_corpus_loader.controller.loader_service.FileLoadError import FileLoadError
from atap_corpus_loader.controller.loader_service.LoaderService import LoaderService


class FileLoaderService(LoaderService):
    """
    Provides methods that handle the logic of loading files and building the DataFrameCorpus object from the loaded
    files.
    Maintains a reference to files loaded as corpus files and files loaded as metadata files.
    """
    def __init__(self, root_directory: str):
        super().__init__()
        self.root_directory: str = self._sanitise_root_dir(root_directory)

    def get_all_files(self, expand_archived: bool) -> list[FileReference]:
        path_iter: Iterator = iglob(f"{self.root_directory}**", recursive=True)
        all_file_refs: list[FileReference] = []
        for path in path_iter:
            if isdir(path):
                continue

            file_refs: list[FileReference] = self.file_ref_factory.get_file_refs_from_path(path, expand_archived)
            all_file_refs.extend(file_refs)

        all_file_refs.sort(key=lambda ref: ref.get_path())

        return all_file_refs

    def add_corpus_files(self, corpus_filepaths: list[str], include_hidden: bool, tqdm_obj: Tqdm):
        for filepath in tqdm_obj(corpus_filepaths, desc="Loading corpus files", unit="files", leave=False):
            file_ref: FileReference = self.file_ref_factory.get_file_ref(filepath)
            if file_ref in self.loaded_corpus_files:
                continue
            if not include_hidden and file_ref.is_hidden():
                continue
            FileLoaderService._check_filepath_permissions(file_ref)

            self.loaded_corpus_files.add(file_ref)
            if file_ref.is_archive():
                try:
                    zip_refs: list[FileReference] = self.file_ref_factory.get_zip_file_refs(filepath)
                except BadZipFile:
                    raise FileLoadError(f"Can't read Zip file as it is malformed: {file_ref.get_filename()}")
                for zip_ref in zip_refs:
                    if not zip_ref.is_hidden() or include_hidden:
                        self.loaded_corpus_files.add(zip_ref)

    def add_meta_files(self, meta_filepaths: list[str], include_hidden: bool, tqdm_obj: Tqdm):
        for filepath in tqdm_obj(meta_filepaths, desc="Loading metadata files", unit="files", leave=False):
            file_ref: FileReference = self.file_ref_factory.get_file_ref(filepath)
            if file_ref in self.loaded_meta_files:
                continue
            if not include_hidden and file_ref.is_hidden():
                continue
            FileLoaderService._check_filepath_permissions(file_ref)

            self.loaded_meta_files.add(file_ref)
            if file_ref.is_archive():
                try:
                    zip_refs: list[FileReference] = self.file_ref_factory.get_zip_file_refs(filepath)
                except BadZipFile:
                    raise FileLoadError(f"Can't read Zip file as it is malformed: {file_ref.get_filename()}")
                for zip_ref in zip_refs:
                    if not zip_ref.is_hidden() or include_hidden:
                        self.loaded_meta_files.add(zip_ref)

    @staticmethod
    def _sanitise_root_dir(root_directory: str) -> str:
        if type(root_directory) is not str:
            raise TypeError(f"root_directory argument: expected string, got {type(root_directory)}")
        sanitised_directory = normpath(root_directory)

        if not sanitised_directory.endswith(sep):
            sanitised_directory += sep

        return sanitised_directory

    @staticmethod
    def _check_filepath_permissions(file_ref: FileReference):
        filepath: str
        if file_ref.is_zipped():
            filepath = file_ref.get_directory_path()
        else:
            filepath = file_ref.get_path()
        if not exists(filepath):
            raise FileLoadError(f"No file found at: {filepath}")
        if not access(filepath, R_OK):
            raise FileLoadError(f"No permissions to read the file at: {filepath}")



import re
from io import BytesIO
from typing import Optional

from panel import Row, Accordion, bind, HSpacer, Column
from panel.layout import Divider
from panel.pane import Markdown
from panel.widgets import Button, TextInput, FileDownload, Select

from atap_corpus_loader.controller import Controller
from atap_corpus_loader.controller.data_objects import ViewCorpusInfo
from atap_corpus_loader.view.gui import AbstractWidget


class CorpusInfoWidget(AbstractWidget):
    def __init__(self, controller: Controller):
        super().__init__()
        self.controller: Controller = controller

        self.corpus_controls = Accordion(toggle=True, width=600,
                                         header_background="#2675C3",
                                         active_header_background="#56AAFC")
        self.corpus_controls.param.watch(self._update_corpus_display, 'active', onlychanged=True)

        self.corpus_display: Markdown = Markdown()

        self.panel = Column(
            self.corpus_controls,
            Row(self.controller.get_export_progress_bar()),
            Divider(),
            self.corpus_display)

    @staticmethod
    def _build_corpus_label(corpus_info: ViewCorpusInfo) -> str:
        name: Optional[str] = corpus_info.name
        if name is None:
            name = " "
        corpus_label: str = f"{name} - {corpus_info.num_rows} document"
        if corpus_info.num_rows != 1:
            corpus_label += 's'
        if corpus_info.parent_name:
            corpus_label += f" - Parent: {corpus_info.parent_name}"

        return corpus_label

    @staticmethod
    def _build_header_markdown_table(corpus_info: ViewCorpusInfo) -> str:
        if len(corpus_info.headers) != len(corpus_info.dtypes):
            return " "

        sanitised_first_row: list[str] = [re.sub(r'([\\\`*_{}[\]()#+\-.!])', r'\\\1', x) for x in corpus_info.first_row_data]
        sanitised_first_row = [re.sub(r'([\n\r\n])', ' ', x) for x in sanitised_first_row]

        header_row = "| Data label " + "| " + " | ".join(corpus_info.headers) + " |"
        spacer_row = "| :-: " * (len(corpus_info.headers) + 1) + "|"
        dtype_row = "| **Datatype** " + "| " + " | ".join(corpus_info.dtypes) + " |"
        data_row = "| **First document** " + "| " + " | ".join(sanitised_first_row) + " |"

        header_table_text = f"**{corpus_info.name}**\n{header_row}\n{spacer_row}\n{dtype_row}\n{data_row}"
        return header_table_text

    def export_corpus(self, corpus_name: str, filetype: str) -> Optional[BytesIO]:
        return self.controller.export_corpus(corpus_name, filetype)

    def rename_corpus(self, corpus_name: str, name: str):
        self.controller.rename_corpus(corpus_name, name)
        self.update_display()

    def delete_corpus(self, corpus_name: str):
        self.controller.delete_corpus(corpus_name)
        self.update_display()

    def _update_corpus_display(self, *_):
        corpora_info: list[ViewCorpusInfo] = self.controller.get_corpora_info()
        if len(corpora_info) == 0:
            self.corpus_display.object = ' '
            self.corpus_display.visible = False
            return

        if len(self.corpus_controls.active) == 0:
            self.corpus_controls.active = [0]

        active_idx: int = self.corpus_controls.active[0]
        if active_idx >= len(corpora_info):
            self.corpus_display.object = ' '
            self.corpus_display.visible = False
            return

        corpus_info: ViewCorpusInfo = corpora_info[active_idx]
        header_markdown_table: str = CorpusInfoWidget._build_header_markdown_table(corpus_info)

        self.corpus_display.object = header_markdown_table
        self.corpus_display.visible = True

    def update_display(self):
        corpora_info: list[ViewCorpusInfo] = self.controller.get_corpora_info()
        export_types: list[str] = self.controller.get_export_types()
        default_filetype: str = export_types[0]

        corpus_controls_objs: list[Row] = []
        for corpus_info in corpora_info:
            label: str = CorpusInfoWidget._build_corpus_label(corpus_info=corpus_info)

            filetype_dropdown = Select(name="Export filetype", options=export_types, value=default_filetype,
                                       width=100, align="center")
            corpus_export_button = FileDownload(
                label=f"Export",
                filename=f"{corpus_info.name}.{default_filetype}",
                callback=bind(self.export_corpus, corpus_name=corpus_info.name, filetype=filetype_dropdown),
                button_type="primary", button_style="solid",
                height=30, width=100,
                align="center")

            def select_fn(event, name=corpus_info.name, export_button=corpus_export_button):
                filename: str = f"{name}.{event.new}"
                export_button.filename = filename
            filetype_dropdown.param.watch(select_fn, ['value'])

            rename_field: TextInput = TextInput(name="Rename corpus", value=corpus_info.name,
                                                align='center', width=150)
            rename_field.param.watch(lambda event, corpus_name=corpus_info.name: self.rename_corpus(corpus_name, event.new), ['value'])
            delete_button: Button = Button(name="Delete corpus", button_type="danger", align='center')
            delete_button.on_click(lambda event, corpus_name=corpus_info.name: self.delete_corpus(corpus_name))

            corpus_control_row = Row(rename_field, filetype_dropdown, corpus_export_button,
                                     HSpacer(), delete_button, name=label)
            corpus_controls_objs.append(corpus_control_row)

        self.corpus_controls.objects = corpus_controls_objs
        self._update_corpus_display()



from .CorpusHeader import CorpusHeader
from .DataType import DataType
from .FileReference import FileReference, ZipFileReference, FileReferenceFactory
from .ViewCorpusInfo import ViewCorpusInfo
from .UniqueNameCorpora import UniqueNameCorpora



from abc import ABC, abstractmethod
from io import BytesIO
from os.path import join, dirname, basename
from typing import Optional
from zipfile import ZipFile, BadZipFile


class FileReference(ABC):
    def __init__(self, path: str):
        self.path: str = path
        self.path_hash: int = hash(self.path)
        self.directory_path: str = dirname(path)
        self.filename: str = basename(path)

        self.filename_no_ext: str
        self.extension: str
        if '.' not in self.filename:
            self.extension = ''
            self.filename_no_ext = self.filename
        else:
            filename_dot_split = self.filename.split('.')
            self.extension = filename_dot_split[-1]
            self.filename_no_ext = '.'.join(filename_dot_split[:-1])

        self.is_ref_archive = self.extension.lower() == 'zip'

    @abstractmethod
    def get_content_buffer(self) -> BytesIO:
        """
        Provides a BytesIO object which contains the contents of the file.
        :return: The BytesIO object containing the file contents
        :rtype: BytesIO
        """
        raise NotImplementedError()

    def __eq__(self, other):
        if not isinstance(other, FileReference):
            return False
        return self.path_hash == other.path_hash

    def __hash__(self):
        return self.path_hash

    def __str__(self):
        return self.get_path()

    def __repr__(self):
        return self.get_path()

    def get_path(self) -> str:
        """
        :return: the path to the file
        :rtype: str
        """
        return self.path

    def get_directory_path(self) -> str:
        """
        :return: the path to the immediate parent directory of the file
        :rtype: str
        """
        return self.directory_path

    def get_filename(self) -> str:
        """
        :return: the filename of the file, including file extension
        :rtype: str
        """
        return self.filename

    def get_filename_no_ext(self) -> str:
        """
        :return: the filename of the file, excluding file extension
        :rtype: str
        """
        return self.filename_no_ext

    def is_hidden(self) -> bool:
        """
        :return: True if the filename begins with a '.', False otherwise
        :rtype: bool
        """
        return self.filename.startswith('.')

    def get_extension(self) -> str:
        """
        :return: the filetype extension of the file (case-sensitive), excluding the '.'.
        If the filename is 'example.txt', this method will return 'txt'.
        :rtype: str
        """
        return self.extension

    def is_archive(self) -> bool:
        """
        Returns True if the file is an archive file (e.g. example.zip), False otherwise.
        Returns False for files within an archive (e.g. example.zip/text.txt)
        :return: True if the file is an archive file (e.g. example.zip), False otherwise.
        :rtype: bool
        """
        return self.is_ref_archive

    @staticmethod
    def is_zipped() -> bool:
        """
        Returns True if the file is within an archive (e.g. example.zip/text.txt), False otherwise.
        Returns False if file is an archive (e.g. example.zip)
        :return: True if the file is within an archive (e.g. example.zip/text.txt), False otherwise.
        :rtype: bool
        """
        return False


class DiskFileReference(FileReference):
    """
    A general purpose object to hold information regarding a specific file in the file system.
    Folder structure is preserved as a path-like string
    """
    def get_content_buffer(self) -> BytesIO:
        """
        Provides a BytesIO object which contains the contents of the file.
        :return: The BytesIO object containing the file contents
        :rtype: BytesIO
        """
        with open(self.get_path(), 'rb') as bytes_f:
            buf = BytesIO(bytes_f.read())
        return buf


class ZipFileReference(FileReference):
    def __init__(self, zip_file: ZipFile, zip_file_path: str, internal_path: str):
        """
        :param zip_file: the ZipFile object corresponding to the zip file that holds this zipped file. This allows multiple zipped files to share the same ZipFile object
        :param zip_file_path: the path to the zip file that holds this zipped file. This can be absolute or relative to the root_directory specified in CorpusLoader
        :param internal_path: the path within the zip file to this zipped file
        """
        super().__init__(join(zip_file_path, internal_path))
        self.zip_file = zip_file
        self.directory_path: str = zip_file_path
        self.internal_directory: str = dirname(internal_path)

    def get_content_buffer(self) -> BytesIO:
        """
        Provides a BytesIO object which contains the contents of the file.
        :return: The BytesIO object containing the file contents
        :rtype: BytesIO
        """
        internal_path = join(self.internal_directory, self.filename)
        with self.zip_file.open(internal_path, force_zip64=True) as zip_f:
            buf = BytesIO(zip_f.read())

        return buf

    @staticmethod
    def is_zipped() -> bool:
        return True


class RemoteFileReference(FileReference):
    def __init__(self, path: str):
        super().__init__(path)
        self.content_buffer: Optional[BytesIO] = None

    def set_content_buffer(self, content_buffer: BytesIO):
        content_buffer.seek(0)
        self.content_buffer = content_buffer

    def get_content_buffer(self) -> BytesIO:
        """
        Provides a BytesIO object which contains the contents of the file.
        :return: The BytesIO object containing the file contents
        :rtype: BytesIO
        """
        if self.content_buffer is None:
            return BytesIO()
        self.content_buffer.seek(0)
        return BytesIO(self.content_buffer.read())


class FileReferenceFactory:
    """
    Implements the Flyweight pattern to mitigate the overhead of re-creating FileReference objects, as the files within
    the file system are expected to change far less frequently than FileReference objects are referred to.
    An add-only cache for FileReference objects is maintained in the form of a dictionary which maps full_path strings
    to the corresponding FileReference object.
    """
    def __init__(self):
        self.file_ref_cache: dict[str, FileReference] = {}

    def clear_cache(self):
        """
        Resets the cache to an empty dictionary
        """
        self.file_ref_cache = {}

    def get_file_refs_from_path(self, path: str, expand_archived: bool) -> list[FileReference]:
        file_refs: list[FileReference]
        curr_file_ref = [self.get_file_ref(path)]
        if path.endswith('.zip'):
            try:
                file_refs = self.get_zip_file_refs(path)
            except BadZipFile:
                return []
            if not expand_archived:
                file_refs = curr_file_ref
        else:
            file_refs = curr_file_ref

        return file_refs

    def get_file_ref(self, path: str) -> FileReference:
        cached_ref: Optional[FileReference] = self.file_ref_cache.get(path)
        if cached_ref is None:
            cached_ref = DiskFileReference(path)
            self.file_ref_cache[path] = cached_ref

        return cached_ref

    def get_zip_file_refs(self, zip_file_path: str) -> list[FileReference]:
        """
        Accepts a zip file and provides a list of FileReference
        objects that correspond to the zipped files within the zip archive.
        :param zip_file_path: the path to the zip archive that holds the files to be listed.
        :return: a list of FileReference objects corresponding to the files within the zip archive
        :rtype: list[FileReference]
        """
        zip_file = ZipFile(zip_file_path)

        file_refs: list[FileReference] = []
        for info in zip_file.infolist():
            if info.is_dir():
                continue
            zip_ref: FileReference = self._get_single_zip_file_ref(zip_file, zip_file_path, info.filename)
            file_refs.append(zip_ref)

        return file_refs

    def _get_single_zip_file_ref(self, zip_file: ZipFile, zip_file_path: str, internal_path: str) -> ZipFileReference:
        full_path: str = join(zip_file_path, internal_path)
        cached_ref: Optional[ZipFileReference] = self.file_ref_cache.get(full_path)
        if cached_ref is None:
            cached_ref = ZipFileReference(zip_file, zip_file_path, internal_path)
            self.file_ref_cache[full_path] = cached_ref

        return cached_ref

    def get_oni_file_ref(self, path: str) -> RemoteFileReference:
        cached_ref: Optional[FileReference] = self.file_ref_cache.get(path)
        if (cached_ref is None) or (type(cached_ref) is not RemoteFileReference):
            cached_ref = RemoteFileReference(path)
            self.file_ref_cache[path] = cached_ref
        cached_ref: RemoteFileReference

        return cached_ref



from abc import abstractmethod, ABC
from datetime import datetime
from typing import Optional

from atap_corpus.corpus.corpus import DataFrameCorpus
from pandas import DataFrame, merge, concat
from panel.widgets import Tqdm

from atap_corpus_loader.controller.data_objects import FileReference, CorpusHeader, FileReferenceFactory
from atap_corpus_loader.controller.loader_service.FileLoadError import FileLoadError
from atap_corpus_loader.controller.loader_service.file_loader_strategy import FileLoaderStrategy, FileLoaderFactory

"""
Some methods in this module utilise Tqdm from the panel library, which breaks the Model-View separation.
This has been done out of necessity for a progress bar for particular operations.
The panel Tqdm is a wrapper for the standard tqdm module and can be replaced if needed.
"""


class LoaderService(ABC):
    def __init__(self):
        self.loaded_corpus_files: set[FileReference] = set()
        self.loaded_meta_files: set[FileReference] = set()
        # Utilise FileReferenceFactory.clear_cache() if memory overhead is raised as an issue.
        self.file_ref_factory: FileReferenceFactory = FileReferenceFactory()

    @abstractmethod
    def get_all_files(self, expand_archived: bool) -> list[FileReference]:
        raise NotImplementedError()

    @abstractmethod
    def add_corpus_files(self, corpus_filepaths: list[str], include_hidden: bool, tqdm_obj: Tqdm):
        raise NotImplementedError()

    @abstractmethod
    def add_meta_files(self, meta_filepaths: list[str], include_hidden: bool, tqdm_obj: Tqdm):
        raise NotImplementedError()

    def is_corpus_loaded(self) -> bool:
        return len(self.loaded_corpus_files) > 0

    def is_meta_loaded(self) -> bool:
        return len(self.loaded_meta_files) > 0

    def get_loaded_corpus_files(self) -> set[FileReference]:
        return set([f for f in self.loaded_corpus_files if not f.is_archive()])

    def get_loaded_meta_files(self) -> set[FileReference]:
        return set(f for f in self.loaded_meta_files if not f.is_archive())

    def remove_corpus_filepath(self, corpus_filepath: str):
        file_ref: FileReference = self.file_ref_factory.get_file_ref(corpus_filepath)
        if file_ref in self.loaded_corpus_files:
            self.loaded_corpus_files.remove(file_ref)

    def remove_meta_filepath(self, meta_filepath: str):
        file_ref: FileReference = self.file_ref_factory.get_file_ref(meta_filepath)
        if file_ref in self.loaded_meta_files:
            self.loaded_meta_files.remove(file_ref)

    def remove_loaded_corpus_files(self):
        self.loaded_corpus_files.clear()

    def remove_loaded_meta_files(self):
        self.loaded_meta_files.clear()

    def remove_all_files(self):
        self.remove_loaded_corpus_files()
        self.remove_loaded_meta_files()

    def get_inferred_corpus_headers(self) -> list[CorpusHeader]:
        return self._get_file_headers(self.get_loaded_corpus_files())

    def get_inferred_meta_headers(self) -> list[CorpusHeader]:
        return self._get_file_headers(self.get_loaded_meta_files())

    def _get_file_headers(self, file_refs: set[FileReference]) -> list[CorpusHeader]:
        headers: Optional[list[CorpusHeader]] = None
        for ref in file_refs:
            file_loader: FileLoaderStrategy = FileLoaderFactory.get_file_loader(ref)
            try:
                path_headers: list[CorpusHeader] = file_loader.get_inferred_headers()
            except UnicodeDecodeError:
                self.remove_corpus_filepath(ref.get_path())
                self.remove_meta_filepath(ref.get_path())
                raise FileLoadError(f"Error loading file at {ref.get_path()}: file is not UTF-8 encoded")
            except Exception as e:
                self.remove_corpus_filepath(ref.get_path())
                self.remove_meta_filepath(ref.get_path())
                raise FileLoadError(f"Error loading file at {ref.get_path()}: {e}")

            if headers is None:
                headers = path_headers
            elif set(headers) != set(path_headers):
                self.remove_corpus_filepath(ref.get_path())
                self.remove_meta_filepath(ref.get_path())
                raise FileLoadError(f"Incompatible data labels in file: {ref.get_path()}")

        if headers is None:
            headers = []

        return headers

    def build_corpus(self, corpus_name: str,
                     corpus_headers: list[CorpusHeader],
                     meta_headers: list[CorpusHeader],
                     text_header: CorpusHeader,
                     corpus_link_header: Optional[CorpusHeader],
                     meta_link_header: Optional[CorpusHeader],
                     tqdm_obj: Tqdm) -> DataFrameCorpus:
        corpus_files: list[FileReference] = sorted(self.get_loaded_corpus_files(), key=lambda f: f.get_path())
        meta_files: list[FileReference] = sorted(self.get_loaded_meta_files(), key=lambda f: f.get_path())

        corpus_df: DataFrame = self._get_concatenated_dataframe(corpus_files, corpus_headers,
                                                                tqdm_obj, "Building corpus")
        meta_df: DataFrame = self._get_concatenated_dataframe(meta_files, meta_headers,
                                                              tqdm_obj, "Building metadata")

        load_corpus: bool = len(corpus_headers) > 0
        load_meta: bool = len(meta_headers) > 0

        final_df: DataFrame
        if load_corpus and load_meta:
            final_df = merge(left=corpus_df, right=meta_df,
                             left_on=corpus_link_header.name, right_on=meta_link_header.name,
                             how='inner', suffixes=(None, '_meta'))
        elif load_corpus:
            final_df = corpus_df
        elif load_meta:
            final_df = meta_df
        else:
            raise ValueError("No corpus headers or metadata headers provided")

        if (corpus_name == '') or (corpus_name is None):
            corpus_name = f"Corpus-{datetime.now()}"

        return DataFrameCorpus.from_dataframe(final_df, text_header.name, corpus_name)

    @staticmethod
    def _get_concatenated_dataframe(file_refs: list[FileReference],
                                    headers: list[CorpusHeader],
                                    tqdm_obj: Tqdm,
                                    loading_msg: str) -> DataFrame:
        if len(file_refs) == 0:
            return DataFrame()
        df_list: list[DataFrame] = []
        for ref in tqdm_obj(file_refs, desc=loading_msg, unit="files", leave=False):
            file_loader: FileLoaderStrategy = FileLoaderFactory.get_file_loader(ref)
            try:
                path_df: DataFrame = file_loader.get_dataframe(headers)
            except UnicodeDecodeError:
                raise FileLoadError(f"Error loading file at {ref.get_path()}: file is not UTF-8 encoded")
            except Exception as e:
                raise FileLoadError(f"Error loading file at {ref.get_path()}: {e}")

            df_list.append(path_df)

        return concat(df_list, ignore_index=True)


from atap_corpus_loader.controller.data_objects import DataType


class CorpusHeader:
    """
    Represents a column before the corpus is built.
    Holds the name, intended datatype, and whether to include the header.
    Also behaves as a Value Object, where the unique value is the name for equality checks and hashing.
    """
    def __init__(self, name: str, datatype: DataType, include: bool = False):
        self.name: str = name
        self.datatype: DataType = datatype
        self.include: bool = include

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"{self.__class__.__name__}: {self.name} [{self.datatype.value}]"

    def __eq__(self, other):
        if type(other) is not CorpusHeader:
            return False
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)



from abc import ABC, abstractmethod

from panel import Row
from panel.layout import Panel


class AbstractWidget(ABC):
    """
    An abstract class for Panel GUI widgets. Provides methods to set widget visibility and update the display of all
    child widgets.
    """
    def __init__(self):
        self.panel: Panel = Row()
        self.children: list[AbstractWidget] = []

    def __panel__(self) -> Panel:
        return self.panel

    def update_displays(self):
        self.update_display()
        for child in self.children:
            child.update_displays()

    def get_visibility(self) -> bool:
        return self.panel.visible

    def set_visibility(self, is_visible: bool):
        self.panel.visible = is_visible

    def toggle_visibility(self):
        self.panel.visible = not self.panel.visible

    @abstractmethod
    def update_display(self):
        raise NotImplementedError()



from io import BytesIO

from docx import Document
from pandas import DataFrame

from atap_corpus_loader.controller.data_objects import CorpusHeader, DataType
from atap_corpus_loader.controller.loader_service.file_loader_strategy.FileLoaderStrategy import FileLoaderStrategy


class DOCXLoaderStrategy(FileLoaderStrategy):
    def get_inferred_headers(self) -> list[CorpusHeader]:
        headers: list[CorpusHeader] = [
            CorpusHeader('document', DataType.TEXT, include=True),
            CorpusHeader('filename', DataType.TEXT),
            CorpusHeader('filepath', DataType.TEXT)
        ]

        return headers

    def get_dataframe(self, headers: list[CorpusHeader]) -> DataFrame:
        included_headers: list[str] = [header.name for header in headers if header.include]
        file_data = {}
        if 'document' in included_headers:
            file_buf: BytesIO = self.file_ref.get_content_buffer()
            docx_doc = Document(file_buf)
            document = '\n'.join([p.text for p in docx_doc.paragraphs])
            file_data['document'] = [document]
        if 'filename' in included_headers:
            file_data['filename'] = [self.file_ref.get_filename_no_ext()]
        if 'filepath' in included_headers:
            file_data['filepath'] = [self.file_ref.get_path()]

        df: DataFrame = DataFrame(file_data, dtype='string')

        return df



from panel import Row, Spacer, Column, HSpacer
from panel.pane import Markdown
from panel.widgets import Button, TextInput, TooltipIcon

from atap_corpus_loader.controller import Controller
from atap_corpus_loader.view import ViewWrapperWidget
from atap_corpus_loader.view.gui import AbstractWidget
from atap_corpus_loader.view.gui.FileSelectorWidget import FileSelectorWidget
from atap_corpus_loader.view.gui.MetaEditorWidget import MetaEditorWidget


class FileLoaderWidget(AbstractWidget):
    LOADER_WIDTH: int = 600

    def __init__(self, view_handler: ViewWrapperWidget, controller: Controller, include_meta_loader: bool):
        super().__init__()
        self.view_handler: ViewWrapperWidget = view_handler
        self.controller: Controller = controller
        
        self.load_as_corpus_button: Button = Button(name='Load as corpus', width=130, button_style='outline', button_type='success')
        self.load_as_corpus_button.on_click(self.load_as_corpus)
        load_corpus_tooltip = self.view_handler.get_tooltip('load_corpus_button')
        load_corpus_tooltip.visible = not include_meta_loader
        self.load_as_corpus_row: Row = Row(self.load_as_corpus_button, load_corpus_tooltip)
        self.load_as_meta_button: Button = Button(name='Load as metadata', width=130, button_style='outline', button_type='success')
        self.load_as_meta_button.on_click(self.load_as_meta)
        load_buttons_tooltip = self.view_handler.get_tooltip('load_buttons')
        self.load_as_meta_row: Row = Row(self.load_as_meta_button, load_buttons_tooltip, visible=include_meta_loader)

        self.unload_selected_button: Button = Button(name="Unload selected", width=120, button_style='outline', button_type='danger', disabled=True, align='end')
        self.unload_selected_button.on_click(self.unload_selected)
        self.unload_all_button: Button = Button(name="Unload all", width=120, button_style='solid', button_type='danger', disabled=True, align='end')
        self.unload_all_button.on_click(self.unload_all)
        self.unload_col: Column = Column(self.unload_selected_button, self.unload_all_button)

        self.loaded_file_info = Markdown()
        
        self.corpus_name_input = TextInput(placeholder='Corpus name', width=150)
        self.build_button: Button = Button(name='Build corpus', button_style='solid', button_type='success')
        self.build_button.on_click(self.build_corpus)
        build_tool_tip: TooltipIcon = self.view_handler.get_tooltip('build_button')
        self.build_button_row: Row = Row(self.corpus_name_input, self.build_button, build_tool_tip, visible=False, align='end')

        self.file_selector = FileSelectorWidget(view_handler, controller, width=self.LOADER_WIDTH)
        self.file_selector.set_button_operation_fn(self._set_button_status_on_operation)
        self.meta_editor = MetaEditorWidget(view_handler, controller)

        self.panel = Row(
            Column(
                self.file_selector,
                Row(Column(
                    Row(self.load_as_corpus_row,
                        self.load_as_meta_row),
                    self.build_button_row
                ),
                    self.loaded_file_info,
                    HSpacer(),
                    self.unload_col,
                    width=self.LOADER_WIDTH),
                Row(self.controller.get_build_progress_bar())
            ),
            Spacer(width=50),
            self.meta_editor)
        self.children = [self.file_selector, self.meta_editor]
        self.update_display()

    def update_display(self):
        self._set_build_buttons_status()
        self.loaded_file_info.object = self.get_loaded_file_info()

    def get_loaded_file_info(self) -> str:
        file_counts: dict[str, int] = self.controller.get_loaded_file_counts()
        count_str: str = ""
        for filetype in file_counts:
            count_str += f"{filetype}: {file_counts[filetype]}\n"

        return count_str

    def _set_build_buttons_status(self, *_):
        files_added: bool = self.controller.is_meta_added() or self.controller.is_corpus_added()
        self.build_button_row.visible = files_added
        self.unload_selected_button.disabled = not files_added
        self.unload_all_button.disabled = not files_added

    def _set_button_status_on_operation(self, curr_loading: bool, *_):
        self.file_selector.selector_widget.disabled = curr_loading
        self.file_selector.show_hidden_files_checkbox.disabled = curr_loading
        self.file_selector.expand_archive_checkbox.disabled = curr_loading
        self.file_selector.select_all_button.disabled = curr_loading
        self.file_selector.file_type_filter.disabled = curr_loading
        self.file_selector.filter_input.disabled = curr_loading

        self.load_as_corpus_button.disabled = curr_loading
        self.load_as_meta_button.disabled = curr_loading
        self.build_button.disabled = curr_loading

    def load_as_corpus(self, *_):
        self._set_button_status_on_operation(curr_loading=True)
        include_hidden: bool = self.file_selector.get_show_hidden_value()
        file_ls: list[str] = self.file_selector.get_selector_value()
        self.view_handler.load_corpus_from_filepaths(file_ls, include_hidden)
        self._set_button_status_on_operation(curr_loading=False)

    def load_as_meta(self, *_):
        self._set_button_status_on_operation(curr_loading=True)
        include_hidden: bool = self.file_selector.get_show_hidden_value()
        file_ls: list[str] = self.file_selector.get_selector_value()
        self.view_handler.load_meta_from_filepaths(file_ls, include_hidden)
        self._set_button_status_on_operation(curr_loading=False)

    def unload_selected(self, *_):
        self._set_button_status_on_operation(curr_loading=True)
        file_ls: list[str] = self.file_selector.get_selector_value()
        self.controller.unload_filepaths(file_ls)
        self.view_handler.update_displays()
        self._set_button_status_on_operation(curr_loading=False)

    def unload_all(self, *_):
        self._set_button_status_on_operation(curr_loading=True)
        self.controller.unload_all()
        self.view_handler.update_displays()
        self._set_button_status_on_operation(curr_loading=False)

    def build_corpus(self, *_):
        self._set_button_status_on_operation(curr_loading=True)
        success: bool = self.view_handler.build_corpus(self.corpus_name_input.value_input)
        if success:
            self.corpus_name_input.value_input = ""
            self.corpus_name_input.value = ""
            self.unload_all()
        self._set_button_status_on_operation(curr_loading=False)



from io import BytesIO

from pandas import DataFrame, read_csv, to_datetime, Series

from atap_corpus_loader.controller.data_objects import CorpusHeader, DataType
from atap_corpus_loader.controller.loader_service.file_loader_strategy.FileLoaderStrategy import FileLoaderStrategy


class CSVLoaderStrategy(FileLoaderStrategy):
    @staticmethod
    def is_datetime_castable(data_series: Series):
        if data_series.dtype != "object":
            return False
        try:
            to_datetime(data_series)
        except (ValueError, TypeError):
            return False
        return True

    @staticmethod
    def _rename_headers(df: DataFrame):
        df.columns = [f'Data_{c}' for c in df.columns.astype(str)]

    @staticmethod
    def _detect_headers(file_buf: BytesIO) -> bool:
        read_rows = 10
        file_buf.seek(0)
        df_no_header = read_csv(file_buf, header=None, nrows=read_rows)
        file_buf.seek(0)
        df_header = read_csv(file_buf, nrows=read_rows)
        file_buf.seek(0)
        return tuple(df_no_header.dtypes) != tuple(df_header.dtypes)

    def get_inferred_headers(self) -> list[CorpusHeader]:
        read_rows = 10
        file_buf: BytesIO = self.file_ref.get_content_buffer()
        contains_header_row: bool = self._detect_headers(file_buf)
        if contains_header_row:
            df = read_csv(file_buf, nrows=read_rows)
        else:
            df = read_csv(file_buf, header=None, nrows=read_rows)
            self._rename_headers(df)
        headers: list[CorpusHeader] = []
        for header_name, dtype_obj in df.dtypes.items():
            if self.is_datetime_castable(df[header_name]):
                headers.append(CorpusHeader(str(header_name), DataType.DATETIME))
                continue

            try:
                dtype = DataType(str(dtype_obj))
            except ValueError:
                dtype = DataType.TEXT
            headers.append(CorpusHeader(str(header_name), dtype))

        return headers

    def get_dataframe(self, headers: list[CorpusHeader]) -> DataFrame:
        file_buf: BytesIO = self.file_ref.get_content_buffer()
        included_headers: list[str] = [header.name for header in headers if header.include]
        if self._detect_headers(file_buf):
            df = read_csv(file_buf, header=0, dtype=object, usecols=included_headers)
        else:
            df = read_csv(file_buf, header=None, dtype=object)
            self._rename_headers(df)
            df = df[included_headers]
        dtypes_applied_df: DataFrame = FileLoaderStrategy._apply_selected_dtypes(df, headers)

        return dtypes_applied_df



from typing import Optional


class ViewCorpusInfo:
    DATA_WIDTH: int = 50

    def __init__(self, name: Optional[str], num_rows: int, parent_name: Optional[str],
                 headers: list[str], dtypes: list[str], first_row_data: list[str]):
        self.name: Optional[str] = name
        self.num_rows: int = num_rows
        self.parent_name: Optional[str] = parent_name
        self.headers: list[str] = headers
        self.dtypes: list[str] = dtypes
        self.first_row_data: list[str] = [x[:ViewCorpusInfo.DATA_WIDTH] for x in first_row_data]

    def __repr__(self):
        return f"ViewCorpusInfo - name: {self.name}"



from io import BytesIO
from xml.etree.ElementTree import parse, Element

from pandas import DataFrame

from atap_corpus_loader.controller.data_objects import CorpusHeader, DataType
from atap_corpus_loader.controller.loader_service.file_loader_strategy import FileLoaderStrategy


class XMLLoaderStrategy(FileLoaderStrategy):
    def _extract_text(self, element: Element) -> str:
        text = element.text or ''
        for child in element:
            text += self._extract_text(child)
            if child.tail:
                text += child.tail

        return text

    def get_inferred_headers(self) -> list[CorpusHeader]:
        headers: list[CorpusHeader] = [
            CorpusHeader('document', DataType.TEXT, include=True),
            CorpusHeader('filename', DataType.TEXT),
            CorpusHeader('filepath', DataType.TEXT)
        ]

        return headers

    def get_dataframe(self, headers: list[CorpusHeader]) -> DataFrame:
        included_headers: list[str] = [header.name for header in headers if header.include]
        file_data = {}
        if 'document' in included_headers:
            file_buf: BytesIO = self.file_ref.get_content_buffer()
            raw_xml_tree = parse(file_buf)
            root = raw_xml_tree.getroot()
            document = self._extract_text(root)

            file_data['document'] = [document]
        if 'filename' in included_headers:
            file_data['filename'] = [self.file_ref.get_filename_no_ext()]
        if 'filepath' in included_headers:
            file_data['filepath'] = [self.file_ref.get_path()]

        df: DataFrame = DataFrame(file_data, dtype='string')

        return df



import pathlib
from glob import iglob
from os.path import sep, isdir, basename, join
from typing import Optional

from panel.widgets import TooltipIcon


class TooltipManager:
    """
    The TooltipManager class is a retrieval interface for TooltipIcon objects.
    The only public method is get_tooltip(), which provides a TooltipIcon object corresponding to the given tooltip_name
    """

    @staticmethod
    def _resolve_dir(dir_path: str) -> str:
        if type(dir_path) is not str:
            raise TypeError(f"expected string argument, got {type(dir_path)}")
        file_parent = pathlib.Path(__file__).parent.resolve()
        sanitised_directory = join(file_parent, dir_path)
        if not sanitised_directory.endswith(sep):
            sanitised_directory += sep

        return sanitised_directory

    @staticmethod
    def _read_markdown_contents() -> dict[str, str]:
        path_pattern: str = f"{TooltipManager.MARKDOWN_DIR}**"
        tooltip_map: dict[str, str] = {}
        for path in iglob(path_pattern, recursive=True):
            if isdir(path):
                continue
            filename_split = basename(path).split('.')
            if len(filename_split) == 1:
                filename = filename_split[0]
            else:
                filename = '.'.join(filename_split[:-1])

            with open(path) as f:
                contents = f.read()
            tooltip_map[filename] = contents

        return tooltip_map

    MARKDOWN_DIR: str = _resolve_dir("./markdown")

    def __init__(self):
        self.tooltip_map: dict[str, str] = self._read_markdown_contents()

    def get_tooltip(self, tooltip_name: str) -> Optional[TooltipIcon]:
        """
        Returns a TooltipIcon object whose text value is the markdown found in the file at MARKDOWN_DIR / tooltip_name + '.md'
        :param tooltip_name: The name of the file in the MARKDOWN_DIR that contains the tooltip markdown (without the '.md' suffix)
        :type tooltip_name: str
        :rtype: TooltipIcon
        """
        text = self.tooltip_map.get(tooltip_name)
        if text is None:
            return None
        return TooltipIcon(value=text, margin=(0, 0))
